from abc import ABC

from .base import Agent


class HighFrequencyAgent(Agent, ABC):
    """High-frequency trading Agent class.

    High-frequency traders are treated as a special agent. Please see xxxx for more details.
    #ToDo

    This class inherits from the :class:`pams.agents.Agent` class and ABC class.
    """

    pass
